/* serial.c   Jed Margolin

  Adapted from SERIALW32.C from Developer's HC08 Bootloader Master
  by Michal Hanak                                   *
*/

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

#define sys_usleep(us) Sleep(us/1000)


// comm funcs

#define FDISCARD 1
#define FDRAIN 2

int init_uart(char* nm, char* speed);
int close_uart(void);
int send_break10(void);
int flush_uart(int out, int in);
int wb(const void* data, unsigned len);
int rb(void* dest, unsigned len);


///////////////////////////////////////////////////////////////
// module vars

static HANDLE port = 0;

///////////////////////////////////////////////////////////////
// open and set tty

int init_uart(char* nm, char* speed)
{
	char fullNm[20];
	DCB dcb;
	DWORD dwErr;
	COMMTIMEOUTS tmo;
	COMSTAT comStat;

	// only port number specified ?
	// prefix with "COM" string
	if(isdigit(*nm))
	{
		strcpy(fullNm, "COM");
		nm = strcat(fullNm, nm);
	}

	port = CreateFile(nm, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING,
						 	    FILE_ATTRIBUTE_NORMAL, NULL);

	if(port == INVALID_HANDLE_VALUE) 
	{
		fprintf(stderr, "Can't open serial port %s\n", nm);
		return -1;
	}

	// init queues
	PurgeComm(port, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);

	dcb.DCBlength = sizeof(dcb);		// sizeof(DCB) 
	dcb.BaudRate = CBR_9600;			// current baud rate 
	dcb.fBinary = TRUE;					// binary mode, no EOF check 
	dcb.fParity = FALSE;				// disable parity checking 
	dcb.fOutxCtsFlow = FALSE;			// CTS output flow control 
	dcb.fOutxDsrFlow = FALSE;			// DSR output flow control 
	dcb.fDtrControl = DTR_CONTROL_ENABLE; // DTR flow control type 
	dcb.fDsrSensitivity = FALSE;		// DSR sensitivity 
	dcb.fTXContinueOnXoff = FALSE;		// XOFF continues Tx 
	dcb.fOutX = FALSE;					// XON/XOFF out flow control 
	dcb.fInX = FALSE;					// XON/XOFF in flow control 
	dcb.fErrorChar = FALSE;				// enable error replacement 
	dcb.fNull = FALSE;					// enable null stripping 
	dcb.fRtsControl = RTS_CONTROL_ENABLE; // RTS flow control 
	dcb.fAbortOnError = FALSE;			// abort reads/writes on error 
	dcb.XonLim = 100;					// transmit XON threshold 
	dcb.XoffLim = 100;					// transmit XOFF threshold 
	dcb.ByteSize = 8;					// number of bits/, 4-8 
	dcb.Parity = NOPARITY;				// 0-4=no,odd,even,mark,space 
	dcb.StopBits = ONESTOPBIT;			// 0,1,2 = 1, 1.5, 2 
	dcb.XonChar = 17;					// Tx and Rx XON acter 
	dcb.XoffChar = 19;					// Tx and Rx XOFF acter 
	dcb.ErrorChar = 0;					// error replacement acter 
	dcb.EofChar = 0;					// end of input acter 
	dcb.EvtChar = 0;					// received event acter 

	// speed string specified ?
	if(speed)
	{
		if(sscanf(speed, "%lu", &dcb.BaudRate) != 1)
		{
			fprintf(stderr, "Invalid speed specified (%s)\n", speed);
			return -1;
		}
	}

	if(!SetCommState(port, &dcb))
	{
		fprintf(stderr, "Can't setup serial port (%s @ %s)\n", nm, speed);
		return -1;
	}

	tmo.ReadIntervalTimeout = 100;		// between two characters
	tmo.ReadTotalTimeoutMultiplier = 1;
	tmo.ReadTotalTimeoutConstant = 500;
	tmo.WriteTotalTimeoutMultiplier = 1;
	tmo.WriteTotalTimeoutConstant = 1000;

	if(!SetCommTimeouts(port, &tmo))
	{
		fprintf(stderr, "Can't set serial port timeouts (%s)\n", nm);
		return -1;
	}
	if(!SetCommMask(port, EV_TXEMPTY))
	{
		fprintf(stderr, "Can't set serial port waitmask (%s)\n", nm);
		return -1;
	}

	// we are ready !
	EscapeCommFunction(port, CLRDTR);
	ClearCommError(port, &dwErr, &comStat);
	return 1;
}


  int close_uart(void)
{
	return CloseHandle(port) ? 0 : -1;
}


  // send 10 zero bits

int send_break10(void)
{
	static const BYTE zero = 0;
	DCB dcb;
	int ret;

	if(!GetCommState(port, &dcb))
		return -1;

	dcb.Parity = EVENPARITY;
	if(!SetCommState(port, &dcb))
		return -1;

	// write hex 0
	ret = wb(&zero, 1);
	if(ret > 0)
		flush_uart(FDRAIN, 0);

	// give him a chance to finish 
	// 1ms is enough, but you never know...
	Sleep(10);

	// revert back to original state
	dcb.Parity = NOPARITY;
	if(!SetCommState(port, &dcb))
		return -1;

	return ret;
}

// flush output out=1=>discard; out>1=>wait_for_tx

int flush_uart(int out, int in)
{
	DWORD dw;

	if(out=FDISCARD) PurgeComm(port, PURGE_TXCLEAR);
	else if(out) WaitCommEvent(port, &dw, NULL);
	if(in) PurgeComm(port, PURGE_RXCLEAR);

	return 0;
}

// write block, return -1=error, or number of bytes written 

int wb(const void* data, unsigned len)
{
	DWORD wrtn;
	BOOL succ = WriteFile(port, data, len, &wrtn, NULL);
	int ret = succ ? wrtn : -1;
	return ret;
}

// read block, return -1=error, or number of bytes read

int rb(void* dest, unsigned len)
{
	DWORD rd;
	BOOL succ = ReadFile(port, dest, len, &rd, NULL);
	int ret = succ ? rd : -1;
	return ret;
}

